from enum import Enum

class HTTPStatusCode(Enum):
    """ Http status code enum class
    """

    OK  = 200
    CREATED = 201
    NO_CONTENT = 204

    BAD_REQUEST = 400
    UNAUTHORIZED = 401
    FORBIDDEN = 403

    NOT_FOUND = 404
    METHOD_NOT_ALLOWED = 405
    NOT_ACCEPTABLE = 406
    CONFLICT = 409
    UNSUPPORTED_TYPE = 415
    UNPROCESABLE_ENTITY = 422

    INTERNAL_SERVER_ERROR = 500